;
; sprite.asm
; NES 64 Sprites source code
; Copyright 2000 Damian Yerrick
; compile with x816 assembler available from
; http://www.zophar.net/
;
; If you are planning on using this code in a Real NES, put it
; in an NROM (e.g. Super Mario Bros.) cartridge.  The iNES
; board number for NROM is 0.
;
;
;
;
;
;
;
;
;
;
;
;
;
;
;
;
;
;
;
; 0000 direct page
; 0080 NT2 replay code direct page area
; 00C0 direct page
; 0100 stack
; 0200 sprite DMA area
; 0300 NT2 replay code main ram area
; 0400 spritedatatables
; 0800 ---
; 8000 replay code
; 9000 music data
; a000 my code
; f000 nametables and other data tables
; fffa interrupt vectors
;

                .opt on
                .mem 8
                .index 8
                .org $c000
                .list

;PPU regs
spraddr = $2003
sprdma = $4014
ppuaddr = $2006
ppudata = $2007


; Work around the 8 sprites per scanline limitation by constantly
; shuffling priorities.  Konami games use this algorithm.
firstsproffset = 16     ; offset to start writing sprites to
                        ; each frame, firstsproffset += 60
cursproffset = 17       ; offset to write next sprite to
                        ; for each sprite written, cursproffset += 68

retraces = $fa          ; (16 bits) the number of retraces
control1 = $fc          ; data read from controllers
control2 = $fd
control3 = $fe
control4 = $ff


sprbuf = $200           ; sprite dma communication area

xpartial = $400
xloc = $440
ypartial = $480
yloc = $4c0
dx = $500
dy = $540
theta = $580
dtheta = $5c0




                .incsrc "replay22.inc" ;Bananmos's NerdTracker II replay code

                .db 13, 10, "music code by bananmos", 13, 10
                .db "everything else copyright 2000 damian yerrick", 13, 10

;
; main()
;
; Nintendo wrote the first part of the init code in _Duck_Hunt_.
; Kevin Horton and Chris Covell played with it before I put it in.
;

resetpoint      cld
                sei
                jsr wait4vbl

                ldx #$00  ;turn off screen display
                stx $2000
                stx $2001
                txa       ;a zero will rest in A for most of the init code

-                               ;clear the NES's CPURAM
                sta $00,x
                sta $100,x
                sta $200,x
                sta $300,x
                sta $400,x
                sta $500,x
                sta $600,x
                sta $700,x
                inx
                bne -

                dex       ;move top of stack to $1ff
                txs

;clear the NES's CPURAM
                

;clear out sprites
                tax
-               lda #$ef
		sta sprbuf,x   ;Y = 241
		inx
		lda #$b6
                sta sprbuf,x   ;index = blank
		inx
                lda #0
		sta sprbuf,x   ;no flip, front, color 0
		inx
                sta sprbuf,x   ;X = 0
		inx
                bne -
                jsr CopySprites
		lda #0

;clear out 4K of nametables
                ldy #$20
                sty ppuaddr
                sta ppuaddr     ;still got the zero... (ppuram address)
                tax
                ldy #2          ;each pass through the x loop will clear 2K
-               sta ppudata
                sta ppudata
                sta ppudata
                sta ppudata
                sta ppudata
                sta ppudata
                sta ppudata
                sta ppudata
                dex
                bne -
                dey
                bne -

;initialize palette
                ldx #$3F
                stx ppuaddr
                sta ppuaddr
                tax
                ldy #$20
-               lda titlepal,x
                sta ppudata
                inx
                dey
                bne -
                tya             ;put that zero back in A

                lda #<obeydata
                sta 0
                lda #>obeydata
                sta 1
                jsr copynametable

                lda #0
                sta ppuaddr
                sta ppuaddr

;initialize sprites
                ldx #$3f
-
                lda #0
                sta xpartial,x
                sta ypartial,x
                lda randnos,x
                sta dx,x
                lda randnos+64,x
                sta dy,x
                txa
                sta theta,x
                lda randnos+128,x
                cmp #224
                bcc +
                eor #%10000000
+
                sta yloc,x
                lda randnos+192,x
                cmp #248
                bcc +
                eor #%10000000
+
                sta xloc,x
                txa
                lsr a
                lsr a
                lsr a
                sec
                sbc #4
                sta dtheta,x
                dex
                bpl -

                lda #0          ;start sprite-ly music
                jsr $8003

mainloop        jsr wait4endvbl
                jsr $8000

                ldy firstsproffset
                tya
                clc
                adc #60
                sta firstsproffset
                ldx #63
-
                lda dx,x
                bpl +
                dec xloc,x
+
                clc
                adc xpartial,x
                sta xpartial,x
                bcc +
                inc xloc,x
+
                lda xloc,x
                cmp #252
                bcc +
                lda #0
+
                sta sprbuf+3,y
                lda xloc,x
                cmp #248
                bcc +
                lda #0
                sec
                sbc dx,x
                sta dx,x
+
                lda dy,x
                bpl +
                dec yloc,x
+
                clc
                adc ypartial,x
                sta ypartial,x
                bcc +
                inc yloc,x
+
                lda yloc,x
                sta sprbuf,y

                cmp #224
                bcc +
                lda #0
                sec
                sbc dy,x
                sta dy,x
+
                lda theta,x
                clc
                adc dtheta,x
                sta theta,x
                and #%11110000
                lsr a
                lsr a
                lsr a
                ora #%10000000
                sta sprbuf+1,y
                lda #%00100001
                sta sprbuf+2,y

                tya
                clc
                adc #68
                tay

                dex
                bpl -

                jsr sort
                jsr wait4vbl

                lda retraces    ;erase
                bne +
                lda retraces+1
                cmp #2
                bne +

                lda #$21
                sta ppuaddr
                lda #$c0
                sta ppuaddr
                ldx #32
                lda #0
-
                sta ppudata
                sta ppudata
                sta ppudata
                sta ppudata
                dex
                bne -
                sta ppuaddr
                sta ppuaddr

+                
                jsr CopySprites

                jmp mainloop










;
; sort
; bubblesorts the sprites.  Very fast if they're already almost sorted.
;

sort
                ldx #0
-
                lda xloc,x      ;compare the sprites
                cmp xloc+1,x
                bcc +           ;if they're correctly sorted, move on
                inx
                cpx #63         
                bcc -
                rts
+                               ;swap them
                jsr swap

                dex             ;move back
                bpl -
                inx
                jmp -
                

;
; swap
; Swaps a pair of sprites x and x+1.
;

swap
                lda xloc,x
                tay
                lda xloc+1,x
                sta xloc,x
                tya
                sta xloc+1,x

                lda xpartial,x
                tay
                lda xpartial+1,x
                sta xpartial,x
                tya
                sta xpartial+1,x

                lda dx,x
                tay
                lda dx+1,x
                sta dx,x
                tya
                sta dx+1,x

                lda yloc,x
                tay
                lda yloc+1,x
                sta yloc,x
                tya
                sta yloc+1,x

                lda ypartial,x
                tay
                lda ypartial+1,x
                sta ypartial,x
                tya
                sta ypartial+1,x

                lda dy,x
                tay
                lda dy+1,x
                sta dy,x
                tya
                sta dy+1,x

                lda theta,x
                tay
                lda theta+1,x
                sta theta,x
                tya
                sta theta+1,x

                lda dtheta,x
                tay
                lda dtheta+1,x
                sta dtheta,x
                tya
                sta dtheta+1,x
                rts


;
; wait4vbl
; waits for vertical blank
;

wait4vbl        bit $2002
                bpl wait4vbl
                rts


;
; wait4endvbl
; waits for end of vertical blank
;

wait4endvbl     bit $2002
                bmi wait4endvbl
                rts


;
; puts
; There is a C-style string in (AAYY). Print it to wherever ppuaddr points.
; Trashes: 0 1
; output: A = 0, Y = string length
;

puts
                sty 0
                sta 1
                ldy #0
-               lda (0),y
                beq +
                sta ppudata
                iny
                bne -
+               rts


;
; puthex
; Writes the number in A to ppudata.
;

puthex          pha
                pha
                lsr a
                lsr a
                lsr a
                lsr a
                tax
                lda hexdigits,x
                sta ppudata
                pla
                and #$0f
                tax
                lda hexdigits,x
                sta ppudata
                pla
                rts
hexdigits       .db "0123456789ABCDEF"


;
; midi2freq
; Converts a MIDI note to a frequency.
; Preserves X, Y, trashes 0. Returns freq in $01:A.
;

midi2freq
                stx 0
                ldx #0
                cmp #72
                bcc midi2freq_less
                sbc #72
                cmp #12
                bcs midi2freq_more
                tax
                lda pitches,x
                ldx #0
                stx 1
                ldx $fe
                rts

midi2freq_less
                ldx #0
                stx 1
-
                inc 1
                adc #12
                cmp #72
                bcc -
                sbc #72
                tax
                lda pitches,x
                pha
                ldx 1
                lda #0
                sta 1
                pla
-               asl
                rol 1
                dex
                bne -
                ldx 0
                rts

pitches         .db 215, 203, 192, 181, 171, 161, 152, 144, 136, 128, 121, 114

midi2freq_more
                ldx #0
                stx 1
-
                inc 1
                sbc #12
                cmp #12
                bcs -
                tax
                lda pitches,x
-               lsr
                dec 1
                bne -
                ldx 0
                rts


;
; copynametable
; Copy a name table from CPUROM (0) to PPURAM $2000.
;

copynametable
                ldy #0
                sty $2000
                sty $2001
                jsr wait4vbl
                lda #$20
                sta ppuaddr
                lda #0
                sta ppuaddr
                ldx #4

-               lda (0),y
                sta ppudata
                iny
                lda (0),y
                sta ppudata
                iny
                lda (0),y
                sta ppudata
                iny
                lda (0),y
                sta ppudata
                iny
                bne -
                inc 1
                dex
                bne -
                jsr wait4endvbl
                jsr wait4vbl
                sty ppuaddr
                sty ppuaddr

                lda #%10100000  ;NMI on VBL; 8x16 sprites; bg patterns 0;
                                ;sprite patterns 0; ppu += 1; nametable $2000
                sta $2000
                lda #%00011110  ;black bg; sprites on fullscreen;
                                ;bg on fullscreen; color display
                sta $2001
                rts


;
; CopySprites
; Copies the sprites from the comm area at $200 to PPU sprite RAM.
; Trashes: nothing
;

CopySprites     pha
                lda #0
                sta spraddr
                nop
                nop
                nop
                lda #>sprbuf
                sta sprdma
                pla
                rts


;
; ReadJPad
; Poll the controllers.
;

ReadJPad
; Read controllers 1 and 3.
                ldx #1
                stx $4016
                dex
                stx $4016
                ldx #8
-               lda $4016
                lsr a
                rol control1
                dex
                bne -
                ldx #8
-               lda $4016
                lsr a
                rol control3
                dex
                bne -

; Read controllers 2 and 4.
                inx
                stx $4017
                dex
                sta $4017
                ldx #8
-               lda $4017
                lsr a
                rol control2
                dex
                bne -
                ldx #8
-               lda $4017
                lsr a
                rol control4
                dex
                bne -
                rts


;
; nmipoint
; Increment retrace.
;

nmipoint
                pha             ;Save all regs.
                txa
                pha
                tya
                pha


;Handle the retrace counter.
                inc retraces
                bne +
                inc retraces+1
+

                pla             ;Restore all regs and exit.
                tay
                pla
                tax
                pla
                rti

;
; DATA TABLES
;

                .pad $fa00      ; enough for four name tables and
                                ; a few other data tables

obeydata        .incbin "sprite.nam"    ;made with nsa
randnos         .incbin "randnos.bin"   ;made with included randnos.c

; palette
titlepal        .db $0f,$00,$10,$30,$0f,$12,$1a,$30,$0f,$1a,$2c,$30,$0f,$12,$14,$30
                .db $0f,$00,$10,$30,$0f,$12,$1a,$30,$0f,$1a,$2c,$30,$0f,$12,$14,$30

irqpoint        rti

;interrupt table
                .pad $fffa
                .dw nmipoint
                .dw resetpoint
                .dw irqpoint

                .incbin "sprite.chr"    ;made with tlayer
